@@ -22,7 +22,11 @@ from .attributes.files import (
     FilesGarbageCollector,
     set_files_root_dir,
 )
-from .attributes.models import ModelAttribute, InlineModelAttribute
+from .attributes.models import (
+    ModelAttribute,
+    InlineModelAttribute,
+    IndexedModelAttribute,
+)
 from .attributes.simples import (
     BooleanAttribute,
     DateAttribute,
@@ -41,17 +45,18 @@ from .exceptions import (
     SherafException,
     ObjectNotFoundException,
     ModelObjectNotFoundException,
-    IndexObjectNotFoundException,
     SameNameForTableException,
     NotConnectedException,
     InvalidFilterException,
+    InvalidIndexException,
     InvalidOrderException,
     QuerySetUnpackException,
     EmptyQuerySetUnpackException,
+    PrimaryKeyException,
 )
-from .indexes import Index
 from .models import (
     AutoModel,
+    AttributeModel,
     IntAutoModel,
     IntIndexedIntAttributesModel,
     IntIndexedModel,
