@@ -1,4 +1,5 @@
-import warnings
+from BTrees.OOBTree import OOBTree
+from sheraf.attributes.indexdetails import IndexDetails
 
 
 READ_MEMOIZATION = False
@@ -28,24 +29,26 @@ class BaseAttribute(object):
                           read or write access. If False, it is created when the model object
                           is created. Default is True.
     :type lazy: :class:`bool`
-    :param lazy_creation: Deprecated in favor of **lazy**.
     :param read_memoization: Whether this attribute should be memoized on read. ``False`` by default.
     :type read_memoization: :class:`bool`
     :param write_memoization: Whether this attribute should be memoized on write. ``True`` by default.
     :type write_memoization: :class:`bool`
 
     When an attribute is memoized, its next reading will not result in a new database access.
+    Attributes:
+    - indexes:    a dictionary of Indexes. The key with value None stands for this attribute's name.
     """
 
+    default_index_mapping = OOBTree
+
     def __init__(
         self,
         default=None,
         key=None,
-        lazy_creation=None,
+        lazy=True,
         read_memoization=None,
         write_memoization=None,
         store_default_value=True,
-        lazy=True,
     ):
         self._default_value = default
         self._default_key = None
@@ -56,19 +59,9 @@ class BaseAttribute(object):
         self.write_memoization = (
             WRITE_MEMOIZATION if write_memoization is None else write_memoization
         )
-
-        if lazy_creation is not None:
-            self.lazy = lazy_creation
-            warnings.warn(
-                "The 'lazy_creation' argument has been deprecated and will be removed in sheraf 0.2. "
-                "Please use 'lazy' instead",
-                DeprecationWarning,
-                stacklevel=2,
-            )
-
-        else:
-            self.lazy = lazy
+        self.lazy = lazy
         self.store_default_value = store_default_value
+        self.indexes = {}
 
     def set_default_key(self, key):
         self._default_key = key
@@ -184,10 +177,74 @@ class BaseAttribute(object):
 
     def delattr(self, parent):
         # Internal. delete attribute from its owner
-        del parent.mapping[self.key(parent)]
+        # TODO: We should remove this try/except
+        try:
+            del parent.mapping[self.key(parent)]
+        except KeyError:
+            pass
 
     def save(self, parent):
         pass
 
     def delete(self, parent):
         pass
+
+    def index(
+        self,
+        unique=False,
+        key=None,
+        values=None,
+        search=None,
+        mapping=None,
+        primary=False,
+    ):
+        """
+        Indexing an attribute allows very fast reading with :func:`~sheraf.queryset.QuerySet.filter` calls.
+
+        :param unique: If the attribute is unique, and two models have the same value for this
+                      attribute, a :class:`~sheraf.exceptions.UniqueIndexException` is raised
+                      when trying to write the second one. Automatically set to :class:`True` if
+                      *primary* is :class:`True`.
+        :param key: The key the index will use. By default, just the attribute name is used.
+        :param values: A callable that takes the current attribute value and returns a collection of values to index. Each generated value will be indexed each time this attribute is edited. It may take time if the generated collection is large. By default, the current attribute raw value is used.
+        :param primary: If true, this will be the default index for the model. `False` by default.
+
+        When indexes are used, **lazy** is disabled.
+
+        >>> class People(sheraf.Model):
+        ...     table = "index_people"
+        ...
+        ...     # Simple indexing
+        ...     name = sheraf.SimpleAttribute().index()
+        ...
+        ...     # Emails can only be owned once
+        ...     email = sheraf.SimpleAttribute().index(unique=True)
+        ...
+        ...     # Indexing people by their decade
+        ...     age = sheraf.SimpleAttribute().index(key="decade", values=lambda age: {age // 10})
+        ...
+        >>> with sheraf.connection(commit=True):
+        ...     m = People.create(name="George Abitbol", email="george@abitbol.com", age=55)
+        ...
+        >>> with sheraf.connection():
+        ...     assert [m] == People.filter(name="George Abitbol")
+        ...     assert [m] == People.filter(decade=5)
+        ...
+        >>> with sheraf.connection():
+        ...     People.create(name="Peter", email="george@abitbol.com", age=35)
+        Traceback (most recent call last):
+            ...
+        UniqueIndexException
+        """
+        self.indexes[key] = IndexDetails(
+            self,
+            unique,
+            key,
+            values,
+            search,
+            mapping or self.default_index_mapping,
+            primary,
+        )
+        self.lazy = False
+
+        return self
