@@ -9,6 +9,7 @@ import sheraf.constants
 from sheraf.exceptions import InvalidFilterException, InvalidOrderException
 
 from collections.abc import Iterable, Sized
+from sheraf.tools.more_itertools import unique_everseen
 
 
 class QuerySet(object):
@@ -56,14 +57,14 @@ class QuerySet(object):
     >>> with sheraf.connection():
     ...    a = Cowboy.all() # returns a QuerySet with all Cowboy
     ...    b = QuerySet([peter, steven, george]) # This is an equivalent custom QuerySet
-    ...    assert set(a) == set(b)
+    ...    assert list(a) == list(b)
 
     :class:`~sheraf.queryset.QuerySet` behave like iterators, and can only be
     consumed once.
 
     >>> with sheraf.connection():
     ...    everybody = Cowboy.all()
-    ...    assert {peter, steven, george} == set(everybody)
+    ...    assert [peter, steven, george] == list(everybody)
     ...    assert [] == everybody
 
     .. note :: :class:`~sheraf.queryset.QuerySet` can be compared against
@@ -117,10 +118,21 @@ class QuerySet(object):
         return "<QuerySet>"
 
     def _model_has_expected_values(self, model):
-        for filter_name, expected_value in self.filters.items():
-            if getattr(model, filter_name) != expected_value:
-                return False
+        for filter_name, expected_value, filter_transformation in self.filters.values():
+            if filter_name in model.indexes():
+                index = model.indexes()[filter_name]
+                if filter_transformation:
+                    if not set(index.details.search_func(expected_value)) & set(
+                        index.details.get_values(model)
+                    ):
+                        return False
+                else:
+                    if expected_value not in index.details.get_values(model):
+                        return False
 
+            elif filter_name in model.attributes:
+                if getattr(model, filter_name) != expected_value:
+                    return False
         return True
 
     def __next__(self):
@@ -128,7 +140,11 @@ class QuerySet(object):
             self._init_iterator()
 
         while True:
-            model = next(self._iterator)
+            try:
+                model = next(self._iterator)
+            except sheraf.exceptions.ModelObjectNotFoundException:
+                continue
+
             if self._model_has_expected_values(model) and (
                 not self._predicate or self._predicate(model)
             ):
@@ -200,16 +216,40 @@ class QuerySet(object):
         # TODO: Avoid to recreate a QuerySet and avoid itertools.islice
         return QuerySet(itertools.islice(self._iterator, start, stop, step))
 
+    def _init_indexed_iterator(self, filter_name, filter_value, filter_transformation):
+        index = self.model.indexes()[filter_name]
+        index_values = (
+            index.details.search_func(filter_value)
+            if filter_transformation
+            else [filter_value]
+        )
+
+        self._iterator = self.model.read_these_valid(**{filter_name: index_values})
+
+        if not index.details.unique:
+            self._iterator = unique_everseen(self._iterator, lambda m: m.identifier)
+
     def _init_default_iterator(self, reverse=False):
-        if self.model:
-            keys = self.model.index_iterkeys(reverse=reverse)
-            self._iterator = self.model.read_these(keys)
-
-        elif reverse:
-            self._iterator = reversed(self._iterable)
-
-        else:
+        if not self.model:
             self._iterator = iter(self._iterable)
+            return
+
+        indexed_filters = (
+            (name, value, transformation)
+            for (name, value, transformation) in self.filters.values()
+            if name in self.model.indexes()
+        )
+
+        for name, value, transformation in indexed_filters:
+            self._init_indexed_iterator(name, value, transformation)
+
+            if self._iterator:
+                return
+
+        if not self._iterator:
+            identifier_index = self.model.indexes()[self.model.primary_key()]
+            keys = identifier_index.iterkeys(reverse)
+            self._iterator = self.model.read_these(keys)
 
     def _init_iterator(self):
         # The default sort order is by ascending identifier
@@ -233,7 +273,8 @@ class QuerySet(object):
         # So we successively sort the list from the less important
         # order to the most important order.
         if self._iterable is None:
-            self._iterable = self.model.read_these(self.model.index_iterkeys())
+            keys = self.model.indexes()[self.model.primary_key()].iterkeys()
+            self._iterable = self.model.read_these(keys)
 
         for attribute, order in reversed(self.orders.items()):
             self._iterable = sorted(
@@ -310,26 +351,70 @@ class QuerySet(object):
         Traceback (most recent call last):
             ...
         sheraf.exceptions.InvalidFilterException: Some filter parameters appeared twice
-        """
-        qs = self.copy()
 
+        .. note::   Filtering on indexed attributes is more performant than filtering on non-indexed attributes. See :func:`~sheraf.attributes.base.BaseAttribute.index`.
+        """
+        return self._filter(False, predicate=predicate, **kwargs)
+
+    def search(self, **kwargs):
+        """
+        Refine a copy of the current :class:`~sheraf.queryset.QuerySet` with further tests.
+
+        This method is very similar to :func:`~sheraf.queryset.QuerySet.filter` except the
+        values it takes are transformed with the same way values are transformed at indexation.
+        TODO: pas très clair
+
+        For instance, if an attribute indexes its values with a lowercase transformation, the
+        :func:`~sheraf.queryset.QuerySet.search` attributes will go through the same
+        transformation. Hence it allows to pass uppercase filter values, while
+        :func:`~sheraf.queryset.QuerySet.filter` does not allow this.
+
+        >>> class MyCustomModel(sheraf.Model):
+        ...     table = "my_custom_model"
+        ...     my_attribute = sheraf.SimpleAttribute().index(
+        ...        values=lambda string: {string.lower()}
+        ...     )
+        ...
+        >>> with sheraf.connection(commit=True):
+        ...     m = MyCustomModel.create(my_attribute="FOO")
+        ...
+        >>> with sheraf.connection():
+        ...     assert [m] == MyCustomModel.search(my_attribute="foo")
+        ...     assert [m] == MyCustomModel.filter(my_attribute="foo")
+        ...
+        ...     assert [m] == MyCustomModel.search(my_attribute="FOO")
+        ...     assert [] == MyCustomModel.filter(my_attribute="FOO")
+        """
+
+        return self._filter(True, **kwargs)
+
+    def _filter(self, transformation, predicate=None, **kwargs):
+        qs = self.copy()
         if self.model:
             for filter_name in kwargs.keys():
-                if filter_name not in self.model.attributes:
+                if (
+                    filter_name not in self.model.attributes
+                    and filter_name not in self.model.indexes()
+                ):
                     raise sheraf.exceptions.InvalidFilterException(
                         "{} has no attribute {}".format(
                             self.model.__name__, filter_name
                         )
                     )
-
-        common_attributes = set(qs.filters) & set(kwargs)
+        kwargs_values = OrderedDict(
+            {
+                filter_name: (filter_name, filter_value, transformation)
+                for filter_name, filter_value in kwargs.items()
+            }
+        )
+        common_attributes = set(qs.filters) & set(kwargs_values)
         invalid_common_attributes = any(
-            key for key in common_attributes if qs.filters[key] != kwargs[key]
+            key for key in common_attributes if qs.filters[key] != kwargs_values[key]
         )
         if invalid_common_attributes:
             raise InvalidFilterException("Some filter parameters appeared twice")
 
-        qs.filters.update(kwargs)
+        qs.filters.update(kwargs_values)
 
         if not qs._predicate:
             qs._predicate = predicate
@@ -388,7 +473,9 @@ class QuerySet(object):
         ... with sheraf.connection():
         ...     assert [george, steven, peter] == Cowboy.all().order(age=sheraf.DESC, name=sheraf.ASC)
 
-        .. note:: As ids are indexed, sorting on ids only is very performant.
+        .. note:: Sorting on indexed attributes is more performant than
+            sorting on other attributes. See
+            :func:`~sheraf.attributes.base.BaseAttribute.index`.
             The less :func:`~sheraf.queryset.QuerySet.order` parameters are
             passed, the better performances will be.
         """
@@ -402,13 +489,6 @@ class QuerySet(object):
                 "Only one 'order' positionnal parameter is allowed."
             )
 
-        if (sys.version_info.major, sys.version_info.minor) < (3, 6) and len(
-            kwargs
-        ) > 1:
-            raise InvalidOrderException(
-                "Only one 'order' named parameter is allowed before python 3.6. Please chain 'order' calls instead."
-            )
-
         qs = self.copy()
 
         if self.model:
